"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.humanFriendlyDescription = humanFriendlyDescription;
exports.visitIR = visitIR;

var _commonTags = require("common-tags");

var _macroAssertions = require("../macro-assertions");

var _checks = require("../utils/checks");

var _IRUtils = require("../type-ir/IRUtils");

function humanFriendlyDescription(ir, state) {
  debugger;
  const {
    typeName,
    instantiatedTypes
  } = state;
  const circularRefs = new Map(typeName && (0, _checks.safeGet)(typeName, instantiatedTypes).circular ? [[typeName, 0]] : []);
  const circularMark = circularRefs.size === 1 ? 0 : undefined;
  return visitIR(ir, { ...state,
    circularRefs,
    circularMark
  });
}

function visitIR(ir, state) {
  let visitorFunction;

  switch (ir.type) {
    case "primitiveType":
      visitorFunction = visitPrimitiveType;
      break;

    case "instantiatedType":
      visitorFunction = visitInstantiatedType;
      break;

    case "objectPattern":
      visitorFunction = visitObjectPattern;
      break;

    case "union":
      visitorFunction = visitUnion;
      break;

    case "tuple":
      visitorFunction = visitTuple;
      break;

    case "literal":
      visitorFunction = visitLiteral;
      break;

    case "builtinType":
      visitorFunction = visitBuiltinType;
      break;

    case "failedIntersection":
      (0, _macroAssertions.throwMaybeAstError)(`found failedIntersection while generating type description. This generally means you have an invalid intersection in your types.`);

    default:
      (0, _macroAssertions.throwUnexpectedError)(`unexpected ir type while generating type description: ${ir.type}`);
  }

  return visitorFunction(ir, state);
}

function visitPrimitiveType(ir) {
  return "type: " + ir.typeName;
}

function stringify(val) {
  return typeof val === "string" ? JSON.stringify(val) : val.toString();
}

function visitLiteral(ir) {
  const {
    value
  } = ir;
  return "value: " + stringify(value);
}

function visitInstantiatedType(ir, state) {
  const {
    instantiatedTypes,
    circularRefs
  } = state;
  const {
    typeName
  } = ir;

  if (circularRefs.has(typeName)) {
    return `ref: #${(0, _checks.safeGet)(typeName, circularRefs)}`;
  }

  const referencedIr = (0, _checks.safeGet)(typeName, instantiatedTypes);
  const {
    value,
    circular
  } = referencedIr;

  if (circular) {
    if (!(0, _IRUtils.isObjectPattern)(value) && !(0, _IRUtils.isUnion)(value)) {
      (0, _macroAssertions.throwUnexpectedError)(`Human-friendly type descriptions are not possible for circular types that are not objects or unions. Please contact me, I didn't know this was possible!`);
    }

    circularRefs.set(typeName, circularRefs.size);
    state.circularMark = circularRefs.size - 1;
  }

  if ((0, _IRUtils.isInstantiatedType)(value)) {
    if (circular) (0, _macroAssertions.throwUnexpectedError)(`found circular instantiated type that only referenced another instantiated type`);
    return visitInstantiatedType(value, state);
  }

  return visitIR(value, state);
}

function stringifyRef(id) {
  return `(id: #${id})`;
}

function visitObjectPattern(ir, state) {
  const {
    properties,
    stringIndexerType,
    numberIndexerType
  } = ir;

  if (properties.length === 0 && !stringIndexerType && !numberIndexerType) {
    return "empty object";
  }

  const {
    circularMark
  } = state;
  let base = `object${circularMark !== undefined ? ` ${stringifyRef(circularMark)}` : ""}`;
  state.circularMark = undefined; // prettier-ignore

  const stringIndexerDescription = stringIndexerType ? (0, _commonTags.html)` 
    where all keys are: 
      ${visitIR(stringIndexerType, state)}
    ` : null; // prettier-ignore

  const numberIndexerDescription = numberIndexerType ? (0, _commonTags.html)`
      where all numeric keys are${stringIndexerDescription !== null ? " also" : ""}:
        ${visitIR(numberIndexerType, state)}
      ` : null;

  if (stringIndexerDescription) {
    // prettier-ignore
    base = (0, _commonTags.html)`
    ${base}
      ${stringIndexerDescription}
    `;
  }

  if (numberIndexerDescription) {
    // prettier-ignore
    base = (0, _commonTags.html)`
    ${base} 
      ${numberIndexerDescription}
    `;
  }

  if (properties.length === 0) return base; // prettier-ignore

  return (0, _commonTags.html)`
    ${base} 
      with properties
        ${properties.map(prop => (0, _commonTags.html)`
              - ${stringify(prop.keyName)}${prop.optional ? ' (optional)' : ""}: 
                  ${visitIR(prop.value, state)}
            `).join("\n")}
   `;
}

function visitUnion(ir, state) {
  const {
    circularMark
  } = state;
  let base = `At least one of`;

  if (circularMark !== undefined) {
    base += " " + stringifyRef(circularMark);
    state.circularMark = undefined;
  }

  base += ":"; // prettier-ignore

  return (0, _commonTags.html)`
    ${base}
      ${ir.childTypes.map(child => `- ${visitIR(child, state)}`).join("\n")}
  `;
}

function visitTuple(ir, state) {
  const {
    childTypes,
    firstOptionalIndex,
    restType
  } = ir;
  const elements = childTypes.map((child, idx) => `- ${idx >= firstOptionalIndex ? "(optional) " : ""} ${visitIR(child, state)}`);

  if (restType) {
    elements.push(`- (rest type) ${visitIR(restType, state)}`);
  } // prettier-ignore


  return (0, _commonTags.html)`
  tuple:
    ${elements.join("\n")}
  `;
}

function visitBuiltinType(ir, state) {
  const {
    typeName,
    elementTypes
  } = ir;
  let description;

  if (typeName === "Array" || typeName === "Set") {
    // prettier-ignore
    description = (0, _commonTags.html)`
    ${typeName} of:
      ${visitIR(elementTypes[0], state)} 
    `;
  } else if (typeName === "Map") {
    // prettier-ignore
    description = (0, _commonTags.html)`
      Map:
        key:
          ${visitIR(elementTypes[0], state)}
        value:
          ${visitIR(elementTypes[1], state)}
    `;
  } else {
    (0, _macroAssertions.throwUnexpectedError)(`unexpected builtin type: ${typeName} while generating description.`);
  }

  return description;
}