"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.getInterfaceIR = getInterfaceIR;
exports.getTypeAliasIR = getTypeAliasIR;
exports.getTypeParameterIR = getTypeParameterIR;
exports.getIR = getIR;

var _core = require("@babel/core");

var _babelPluginMacros = require("babel-plugin-macros");

var _macroAssertions = require("../macro-assertions");

var _checks = require("../utils/checks");

var _IRUtils = require("./IRUtils");

function assertTypeAnnotation(node) {
  if (node === null) {
    (0, _macroAssertions.throwMaybeAstError)(`type annotation was null`);
  }
}

/**
 * Code is written in a defensive style. Example:
 * if (cond) {
 *    return value
 * } else {
 *   throw Error(...)
 * }
 *
 * Here, the else statement is not needed, but we keep it.
 * Otherwise, we have the following scenario:
 * if (cond) {
 *    if (cond2) return value
 *    // whoops forgot to handle case when cond2 is false!
 * }
 * // now we falsely throw this error instead
 * throw Error(...)
 */
// Interfaces and type aliases need their own functions to generate IR because
// 1. they are top level (cannot be nested)
// 2. they can have type parameters
function getInterfaceIR(node, externalTypes) {
  const typeParameterInfo = processGenericTypeParameters(node.typeParameters, externalTypes);
  const interface_ = {
    type: "interface",
    ...typeParameterInfo,
    body: getBodyIR(node.body.body, {
      externalTypes,
      typeParameterNames: typeParameterInfo.typeParameterNames,
      parent: node
    })
  };
  return interface_;
} // This function seems like a duplicate of the interface function
// but separation is good in case we implement "extends" in the future


function getTypeAliasIR(node, externalTypes) {
  const typeParameterInfo = processGenericTypeParameters(node.typeParameters, externalTypes);
  const alias = {
    type: "alias",
    ...typeParameterInfo,
    value: getIR(node.typeAnnotation, {
      externalTypes,
      typeParameterNames: typeParameterInfo.typeParameterNames,
      parent: node
    })
  };
  return alias;
}

function processGenericTypeParameters(node, externalTypes) {
  const typeParameterNames = [];
  const typeParameterDefaults = [];
  const typeParameterInfo = {
    typeParameterNames,
    typeParameterDefaults
  };

  if (node === undefined || node === null) {
    return { ...typeParameterInfo,
      typeParametersLength: 0
    };
  }

  for (const param of node.params) {
    // we don't handle type constraints because
    // the macro is not a type checker
    if (param.default) {
      typeParameterDefaults.push(getIR(param.default, {
        externalTypes,
        typeParameterNames,
        parent: node
      }));
    }

    typeParameterNames.push(param.name);
  }

  return { ...typeParameterInfo,
    typeParametersLength: typeParameterNames.length
  };
} // parse the body of a Typescript interface or object pattern


function getBodyIR(elements, state) {
  // https://stackoverflow.com/questions/53276792/define-a-list-of-optional-keys-for-typescript-record
  const indexSignatures = {};
  const propertySignatures = [];

  for (const element of elements) {
    if (_core.types.isTSIndexSignature(element)) {
      var _element$parameters$;

      //                 ˅˅˅˅˅˅˅ is value type annotation (value)
      // {[key: string]: number}
      //   ˄˄˄˄˄˄˄˄˄˄˄˄  is key type annotation (key)
      // The only valid keys are "string" and "number" and
      // keys cannot be optional in index types and key names don't matter
      const keyTypeAnnotation = (_element$parameters$ = element.parameters[0]) === null || _element$parameters$ === void 0 ? void 0 : _element$parameters$.typeAnnotation;

      if (_core.types.isTSTypeAnnotation(keyTypeAnnotation)) {
        const indexType = keyTypeAnnotation.typeAnnotation.type;
        let keyType;

        switch (indexType) {
          case "TSNumberKeyword":
            keyType = "number";
            break;

          case "TSStringKeyword":
            keyType = "string";
            break;

          default:
            (0, _macroAssertions.throwMaybeAstError)(`indexType had an unexpected value: ${indexType}`);
        }

        assertTypeAnnotation(element.typeAnnotation);
        indexSignatures[keyType] = getIR(element.typeAnnotation.typeAnnotation, state);
      } else {
        (0, _macroAssertions.throwMaybeAstError)(`keyTypeAnnotation had unexpected value: ${keyTypeAnnotation}`);
      }
    } else if (_core.types.isTSPropertySignature(element)) {
      const {
        key
      } = element;

      if (_core.types.isIdentifier(key) || _core.types.isStringLiteral(key) || _core.types.isNumericLiteral(key)) {
        const keyName = _core.types.isIdentifier(key) ? key.name : key.value;

        if (typeof keyName === "string" || typeof keyName === "number") {
          const optional = Boolean(element.optional);
          assertTypeAnnotation(element.typeAnnotation);
          propertySignatures.push({
            type: "propertySignature",
            keyName,
            optional,
            value: getIR(element.typeAnnotation.typeAnnotation, state)
          });
        } else {
          (0, _macroAssertions.throwMaybeAstError)(`property signature key ${keyName} had unexpected type ${typeof keyName}`);
        }
      } else {
        (0, _macroAssertions.throwMaybeAstError)(`property signature had unexpected key type: ${key.type}`);
      }
    } else if (_core.types.isTSMethodSignature(element)) {
      // TODO: Have switch that supports method signatures as functions (without type args)
      throw new _babelPluginMacros.MacroError(`Method signatures in interfaces and type literals are not supported`);
    } else {
      (0, _macroAssertions.throwUnexpectedError)(`unexpected signature type: ${element.type}`);
    }
  }

  const [n, s] = [indexSignatures.number, indexSignatures.string];
  const objectPattern = {
    type: "objectPattern",
    properties: propertySignatures,
    ...(n && {
      numberIndexerType: n
    }),
    ...(s && {
      stringIndexerType: s
    })
  };
  return objectPattern;
}

function getTypeParameterIR(node) {
  // Called from createValidator, at this point all type registering has finished
  return getIR(node, {
    // registering of external types has finished
    externalTypes: new Set(),
    // no generic parameters, like T, in a type instantion
    // createValidator<Foo<T>>() is incoherent
    typeParameterNames: [],
    parent: node
  });
}

function getIR(node, oldState) {
  const state = { ...oldState,
    parent: node
  };

  if (_core.types.isTSUnionType(node)) {
    const children = [];

    for (const childType of node.types) {
      children.push(getIR(childType, state));
    }

    if ((0, _checks.hasAtLeast2Elements)(children)) {
      // Have an unnecessary variable instead of return {...} as Union
      // because the latter gives worse type checking
      const union = {
        type: "union",
        childTypes: children
      };
      return union;
    } else {
      (0, _macroAssertions.throwMaybeAstError)(`union type had ${children.length}, which is not possible`);
    }
  } else if (_core.types.isTSParenthesizedType(node)) {
    const childType = node.typeAnnotation;
    return getIR(childType, state);
  } else if (_core.types.isTSIntersectionType(node)) {
    const childTypes = [];

    for (const childType of node.types) {
      childTypes.push(getIR(childType, state));
    }

    if ((0, _checks.hasAtLeast2Elements)(childTypes)) {
      const intersectionType = {
        type: "intersection",
        childTypes: childTypes
      };
      return intersectionType;
    } else {
      (0, _macroAssertions.throwMaybeAstError)(`intersection type had ${childTypes.length}, which is not possible`);
    }
  } else if (_core.types.isTSTupleType(node)) {
    let firstOptionalIndex = -1;
    let restType = null;
    const children = [];
    const {
      elementTypes
    } = node;
    const length = elementTypes.length;

    for (let i = 0; i < length; ++i) {
      const child = elementTypes[i];

      if (_core.types.isTSOptionalType(child)) {
        if (firstOptionalIndex === -1) firstOptionalIndex = i;
        children.push(getIR(child.typeAnnotation, state));
      } else if (_core.types.isTSRestType(child)) {
        if (i !== length - 1) {
          (0, _macroAssertions.throwMaybeAstError)(`rest element was not last type in tuple type because it had index ${i}`);
        }

        const ir = getIR(child.typeAnnotation, state);
        (0, _IRUtils.assertArray)(ir);
        restType = ir;
      } else {
        children.push(getIR(child, state));
      }
    }

    if (firstOptionalIndex === -1) {
      firstOptionalIndex = restType ? length - 1 : length;
    }

    if (firstOptionalIndex === -1) firstOptionalIndex = length;
    const restTypeElement = restType ? restType.typeParameters[0] : undefined;
    const tuple = {
      type: "tuple",
      childTypes: children,
      firstOptionalIndex,
      ...(restTypeElement && {
        restType: restTypeElement
      }),
      undefinedOptionals: true
    };
    return tuple;
  } else if (_core.types.isTSArrayType(node)) {
    // convert Foo[] to Array<Foo>
    // this allows the hoisting pass to identify:
    // val and val2 as the same type in:
    // interface Bar { val: Foo[], val2: Foo[] }
    const arrayType = {
      type: "type",
      typeName: "Array",
      typeParameters: [getIR(node.elementType, state)]
    };
    return arrayType;
  } else if (_core.types.isTSTypeLiteral(node)) {
    // We don't need to worry that the object pattern has references to a generic parameter
    // because this is only possible if it belongs to an interface
    return getBodyIR(node.members, state);
  } else if (_core.types.isTSTypeReference(node)) {
    const typeParameters = [];

    if (node.typeParameters) {
      for (const param of node.typeParameters.params) {
        typeParameters.push(getIR(param, state));
      }
    }

    if (_core.types.isTSQualifiedName(node.typeName)) {
      // TODO: When does a TSQualifiedName pop-up?
      (0, _macroAssertions.throwUnexpectedError)(`typeName was a TSQualifiedName instead of Identifier.`);
    }

    const {
      typeParameterNames,
      externalTypes
    } = state;
    let typeName = node.typeName.name;
    const idx = typeParameterNames.indexOf(typeName);

    if (idx !== -1) {
      if (typeParameters.length > 0) {
        (0, _macroAssertions.throwMaybeAstError)(`Generic parameter ${typeName} had type arguments`);
      }

      const genericType = {
        type: "genericType",
        typeParameterIndex: idx
      };
      return genericType;
    } // Readonly types are indistinguishable from their normal variants at runtime


    if (typeName === "ReadonlyArray") typeName = "Array";else if (typeName === "ReadonlyMap") typeName = "Map";else if (typeName === "ReadonlySet") typeName = "Set"; // It's only an external type if it's not referencing
    // a generic parameter to the parent interface

    externalTypes.add(typeName);
    const withoutTypeParameters = {
      type: "type",
      typeName
    };

    if ((0, _checks.hasAtLeast1Element)(typeParameters)) {
      const type = { ...withoutTypeParameters,
        typeParameters
      };
      return type;
    }

    return withoutTypeParameters;
  } else if (_core.types.isTSLiteralType(node)) {
    const value = node.literal.value;
    const literal = {
      type: "literal",
      value
    };
    return literal;
  } else if (_core.types.isTSNumberKeyword(node) || _core.types.isTSBigIntKeyword(node) || _core.types.isTSStringKeyword(node) || _core.types.isTSBooleanKeyword(node) || _core.types.isTSObjectKeyword(node) || _core.types.isTSNullKeyword(node) || _core.types.isTSUndefinedKeyword(node) || _core.types.isTSAnyKeyword(node) || _core.types.isTSUnknownKeyword(node)) {
    const {
      type
    } = node; // type is "TSNumberKeyword", "TSStringKeyword", etc.

    const builtinTypeName = type.slice("TS".length, -"Keyword".length).toLowerCase();
    (0, _IRUtils.assertPrimitiveTypeName)(builtinTypeName);
    const builtinType = {
      type: "primitiveType",
      typeName: builtinTypeName
    };
    return builtinType;
  } else if (_core.types.isTSMappedType(node)) {
    throw new _babelPluginMacros.MacroError(`${node.type} types are not supported. File an issue with the developer if you want this.`);
  } else {
    (0, _macroAssertions.throwUnexpectedError)(`${node.type} was not expected`);
  }
}