"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.flatten = flatten;
exports.generateBooleanExpr = generateBooleanExpr;
exports.default = void 0;

var _macroAssertions = require("../../macro-assertions");

var _checks = require("../../utils/checks");

var _utils = require("./utils");

var _IRUtils = require("../IRUtils");

/**
 * We want to simplify:
 * 1. A | (B | C) to A | B | C
 * 2. A & (B | C) to A & B | A & C
 *
 * case #1 is important because by simplifying nested unions,
 * we know whether "undefined" is part of an union. If it is,
 * then we have to use Object.prototype.hasOwnProperty in the codegen,
 * otherwise, we don't.
 *
 * case #2 is important because when we generate code, we don't know
 * how to compute the type of A & (B | C), but we can
 * compute the types of A & B and A & C
 *
 * Type IR doesn't have parenthesis, instead it using nesting to convey
 * priority/what parenthesis would convey.
 * As such, an expression like A & (B | C) is converted into (pseudo type IR):
 *
 * intersection: [
 *  A,
 *  union: [B, C]
 *  ]
 *
 * we need to expand this into
 *
 * union: [
 *   intersection: [A, B],
 *   intersection: [B, C]
 * ]
 *
 * First, we convert the type into a javascript boolean expression.
 * A & (B | C) --> A && (B || C)
 *
 * Then we evaluate it with a truth table (0 = false, 1 = true)
 * | A | B | C | Result |
 * |---|---|---|--------|
 * | 0 | 0 | 0 | 0      |
 * | 0 | 0 | 1 | 0      |
 * | 0 | 1 | 0 | 0      |
 * | 0 | 1 | 1 | 0      |
 * | 1 | 0 | 0 | 0      |
 * | 1 | 0 | 1 | 1      |
 * | 1 | 1 | 0 | 1      |
 * | 1 | 1 | 1 | 1      |
 *
 * The resulting valid configurations are:
 * A & C, A & B, A & B & C
 *
 * A & B and A & C are the expected result of
 * expanding A & (B | C).
 *
 * A & B & C is a subtype of both, so there's no harm
 * in having it.
 */
// This is only exported for unit testing purposes
function flatten(ir) {
  const map = {};
  const state = {
    totalNumVars: 0,
    varToValue: map,
    typeNameToVar: new Map(),
    duplicatedType: false
  }; // Generate the boolean expression where the types are substituted with variables

  const expression = generateBooleanExpr(ir, state);
  const withoutOuterParens = expression.slice(1, -1);
  const hasParens = /[()]/.test(withoutOuterParens);

  if (!hasParens && !state.duplicatedType) {
    // TODO: Test this
    // there are no nested unions / intersections
    // AND there are no duplicate types (string & string)
    // so no optimization/simplification can be done
    return ir;
  }

  const hasAnd = /\&\&/.test(withoutOuterParens); // there are no intersections, so we can just remove all parenthesis
  // (it's only unions)

  const simpleUnion = !hasAnd;
  let bitConfigs = []; // There are 2 ^ (number of vars) possible boolean configurations. Test all of them.

  outer: for (let i = 0; i < Math.pow(2, state.totalNumVars); ++i) {
    let copy = expression.slice();
    let bitsSet = 0;
    const trueVars = [];

    for (let j = 0; j < state.totalNumVars; ++j) {
      const jthBitIsSet = (i & 1 << j) !== 0;

      if (jthBitIsSet) {
        trueVars.push(j);
        bitsSet++;
      }

      if (bitsSet > 1 && simpleUnion) {
        continue outer;
      }

      copy = copy.replace(new RegExp(`v${j}`, "g"), jthBitIsSet ? "true" : "false");
    }

    const res = eval(copy);

    if (res) {
      bitConfigs.push(trueVars);
    }
  }

  if (bitConfigs.length === 0) {
    (0, _macroAssertions.throwMaybeAstError)(`for type: ${JSON.stringify(ir)}, could not find valid type configuration`);
  } // We have found all possible minimal true configurations
  // now map the vars back to their original types and generate
  // the new type IR


  const irConfigs = [];

  for (const config of bitConfigs) {
    const childTypes = [];

    for (const var_ of config) {
      const ir = map[var_];

      if (ir === undefined) {
        (0, _macroAssertions.throwUnexpectedError)(`could not de-serialize idx ${var_} back into type. The map was: ${JSON.stringify(map, null, 2)}`);
      }

      childTypes.push(ir);
    }

    if (childTypes.length === 1) {
      /**
       * If there's only one element in a valid config. For example:
       * type T = A | B | C. The valid configs are {A, B, C}
       * then don't create an intersection type
       */
      irConfigs.push(childTypes[0]);
    } else if ((0, _checks.hasAtLeast2Elements)(childTypes)) {
      const intersection = {
        type: "intersection",
        childTypes
      };
      irConfigs.push(intersection);
    } else {
      (0, _macroAssertions.throwUnexpectedError)(`valid configuration had less than 1 true type: ${childTypes.length}`);
    }
  }

  if (irConfigs.length === 1) {
    const soleConfig = irConfigs[0]; // TODO: We should probably do some checking to ensure
    // the soleConfig is a valid type.
    // 3 cases where a union/intersection reduces to a single configuration:
    // A & B = A & B
    // A | A | A = A
    // A & A = A
    // The first case shouldn't occur anyway because we have an optimization
    // to ignore simple unions/intersections. So the code wouldn't reach this
    // point.

    return soleConfig;
  } else if ((0, _checks.hasAtLeast2Elements)(irConfigs)) {
    // A & (B | C) = A & B | A & C
    // create an union out of the valid type configurations
    const union = {
      type: "union",
      childTypes: irConfigs
    };
    return union;
  } else {
    (0, _macroAssertions.throwUnexpectedError)(`irConfigs had length: ${irConfigs.length} even though bitConfigs had length: ${bitConfigs.length}`);
  }
}

/**
 * Converts a Typescript type to a Javascript boolean expression and
 * adds the types to a mapping of variable name in expression --> type
 */
// only exported for testing purposes
function generateBooleanExpr(ir, state) {
  const irIsUnion = (0, _IRUtils.isUnion)(ir);
  const {
    childTypes
  } = ir;
  let expr = "(";

  for (let i = 0; i < childTypes.length; ++i) {
    const type = childTypes[i];
    let childExpr = "";

    if ((0, _IRUtils.isIntersectionOrUnion)(type)) {
      childExpr = generateBooleanExpr(type, state);
    } else {
      const {
        varToValue,
        typeNameToVar
      } = state;
      const typeName = (0, _IRUtils.isPrimitive)(type) ? type.typeName : (0, _IRUtils.isType)(type) || (0, _IRUtils.isLiteral)(type) ? (0, _utils.getTypeKey)(type) : null;
      let var_ = null;

      if (typeName !== null) {
        const existingVar = typeNameToVar.get(typeName);

        if (existingVar !== undefined) {
          state.duplicatedType = true;
          var_ = existingVar;
        } else typeNameToVar.set(typeName, state.totalNumVars);
      }

      var_ = var_ !== null ? var_ : state.totalNumVars++;
      varToValue[var_] = type;
      childExpr = `v${var_}`;
    }

    expr += i === 0 ? childExpr : `${irIsUnion ? " ||" : " &&"} ${childExpr}`;
  }

  return expr + ")";
}

function flattenUnionsAndIntersections(ir) {
  return (0, _utils.traverse)(ir, _IRUtils.isIntersectionOrUnion, cur => {
    return flatten(cur);
  });
}

var _default = flattenUnionsAndIntersections;
exports.default = _default;