"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.traverse = traverse;
exports.applyTypeParameters = applyTypeParameters;
exports.getTypeKey = getTypeKey;
exports.getTypeInfo = getTypeInfo;
exports.getInstantiatedType = getInstantiatedType;

var _babelPluginMacros = require("babel-plugin-macros");

var _macroAssertions = require("../../macro-assertions");

var _fastCopy = _interopRequireDefault(require("fast-copy"));

var _stringify = require("../../utils/stringify");

var u = _interopRequireWildcard(require("../IRUtils"));

function _getRequireWildcardCache() { if (typeof WeakMap !== "function") return null; var cache = new WeakMap(); _getRequireWildcardCache = function () { return cache; }; return cache; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Replace all objects that in ir that match
 * shouldProcess with the result of calling process
 * on them. Returns a copy.
 */
function traverse(ir, shouldProcess, process, skip) {
  function callSkip(x) {
    if (skip) return skip(x);
    return false;
  }

  function helper(current) {
    for (const [k, v] of Object.entries(current)) {
      if (typeof v !== "object" || v === null) continue;

      if (callSkip(v)) {
        continue;
      }

      if (shouldProcess(v)) {
        // @ts-ignore
        current[k] = process(v);
      } else if (Array.isArray(v)) {
        for (let i = 0; i < v.length; ++i) {
          const element = v[i];
          if (callSkip(element)) continue;

          if (shouldProcess(element)) {
            v[i] = process(element);
          } else helper(element);
        }
      } else helper(v);
    }
  }

  const copy = (0, _fastCopy.default)(ir);

  if (callSkip(copy)) {
    return copy;
  }

  if (shouldProcess(copy)) {
    return process(copy);
  }

  helper(copy);
  return copy;
}

function replaceTypeParameters(ir, resolvedParameterValues, currentTypeParameterIndex = resolvedParameterValues.length) {
  return traverse(ir, u.isGenericType, typeParameterRef => {
    const {
      typeParameterIndex
    } = typeParameterRef;

    if (typeParameterIndex >= currentTypeParameterIndex) {
      // TODO: Right now this error isn't thrown in Foo<X = Z, Z> {}
      // because Z wil be parsed as a type reference instead of a type parameter
      throw new _babelPluginMacros.MacroError(_macroAssertions.Errors.InvalidTypeParameterReference(currentTypeParameterIndex, typeParameterIndex));
    }

    return resolvedParameterValues[typeParameterIndex];
  });
}

function applyTypeParameters(target, typeName, providedTypeParameters) {
  const {
    typeParameterDefaults,
    typeParametersLength
  } = target;

  if (typeParametersLength < providedTypeParameters.length) {
    throw new _babelPluginMacros.MacroError(_macroAssertions.Errors.TooManyTypeParameters(typeName, providedTypeParameters.length, typeParametersLength));
  }

  const requiredTypeParameters = typeParametersLength - typeParameterDefaults.length;

  if (requiredTypeParameters > providedTypeParameters.length) {
    throw new _babelPluginMacros.MacroError(_macroAssertions.Errors.NotEnoughTypeParameters(typeName, providedTypeParameters.length, requiredTypeParameters));
  }

  const resolvedParameterValues = providedTypeParameters;

  for (let i = providedTypeParameters.length; i < typeParametersLength; ++i) {
    const instantiatedDefaultValue = replaceTypeParameters(typeParameterDefaults[i - requiredTypeParameters], resolvedParameterValues, i);
    resolvedParameterValues.push(instantiatedDefaultValue);
  }

  return replaceTypeParameters((0, u.isTypeAlias)(target) ? target.value : (0, u.isBuiltinType)(target) ? target : target.body, resolvedParameterValues);
}

function getTypeKey(type) {
  if (u.isType(type)) {
    const {
      typeName,
      typeParameters = []
    } = type;
    if (typeParameters.length === 0) return typeName;
    return typeName + (0, _stringify.deterministicStringify)(typeParameters);
  } else {
    const {
      value
    } = type;
    return `LITERAL?#$<>-${JSON.stringify(value)}`;
  }
}

function getTypeInfo(ir) {
  const {
    type
  } = ir; // TODO: Refactor this switch stmt

  switch (type) {
    case "instantiatedType":
      (0, _macroAssertions.throwUnexpectedError)(`did not expect ${type}, it have been retrieved before calling this method`);

    case "type":
    case "genericType":
      (0, _macroAssertions.throwUnexpectedError)(`did not expect ${type}, it should have been removed during the instantiation pass`);

    case "union":
    case "intersection":
      (0, _macroAssertions.throwUnexpectedError)(`did not expect ${type}, it should have been removed during the flattening pass`);

    case "interface":
    case "alias":
      (0, _macroAssertions.throwUnexpectedError)(`did not expect ${type}, it should only appear as a top level declaration`);

    case "propertySignature":
      (0, _macroAssertions.throwUnexpectedError)(`did not expect ${type}, it should only appear inside an object pattern`);

    default:
      if ((0, u.isBuiltinType)(ir)) {
        const {
          typeName
        } = ir;
        return {
          disjointType: typeName
        };
      }

      if ((0, u.isTuple)(ir)) {
        return {
          disjointType: "Array"
        };
      }

      if ((0, u.isPrimitive)(ir)) {
        switch (ir.typeName) {
          case "bigInt":
            throw new _babelPluginMacros.MacroError("bigInts are not supported yet. Contact the developer if you want to increase their priority.");

          case "any":
          case "unknown":
            return {
              isAnything: true
            };

          default:
            return {
              disjointType: ir.typeName
            };
        }
      }

      if ((0, u.isLiteral)(ir)) {
        const literalType = typeof ir.value;

        switch (literalType) {
          case "boolean":
          case "number":
          case "string":
            return {
              disjointType: literalType,
              literalValue: ir.value
            };

          case "bigint":
          case "symbol":
            throw new _babelPluginMacros.MacroError(`${literalType}s are not supported yet. Contact the developer if you want to increase their priority`);

          default:
            (0, _macroAssertions.throwUnexpectedError)(`unexpected literal type: ${literalType} for value ${ir.value}`);
        }
      }

      if (u.isObjectPattern(ir)) {
        return {
          disjointType: "object"
        };
      }

      (0, _macroAssertions.throwUnexpectedError)(`Failed to get hierarchy info for: ${JSON.stringify(ir, null, 2)}`);
  }
}

function getInstantiatedType(typeName, instantiatedTypes) {
  const instantiated = instantiatedTypes.get(typeName);

  if (instantiated === undefined) {
    (0, _macroAssertions.throwUnexpectedError)(`failed to retrieve instantiated type: ${typeName}`);
  }

  return instantiated.value;
}